@@ -247,6 +247,31 @@ static inline void drbd_generic_make_request(struct drbd_conf *mdev,
 		generic_make_request(bio);
 }
 
+/* see 7eaceac block: remove per-queue plugging */
+#ifdef blk_queue_plugged
+static inline void drbd_plug_device(struct drbd_conf *mdev)
+{
+	struct request_queue *q;
+	q = bdev_get_queue(mdev->this_bdev);
+
+	spin_lock_irq(q->queue_lock);
+
+/* XXX the check on !blk_queue_plugged is redundant,
+ * implicitly checked in blk_plug_device */
+
+	if (!blk_queue_plugged(q)) {
+		blk_plug_device(q);
+		del_timer(&q->unplug_timer);
+		/* unplugging should not happen automatically... */
+	}
+	spin_unlock_irq(q->queue_lock);
+}
+#else
+static inline void drbd_plug_device(struct drbd_conf *mdev)
+{
+}
+#endif
+
 static inline int drbd_backing_bdev_events(struct drbd_conf *mdev)
 {
 	struct gendisk *disk = mdev->ldev->backing_bdev->bd_contains->bd_disk;
@@ -405,6 +430,24 @@ static inline int crypto_hash_final(struct hash_desc *desc, u8 *out)
 
 #endif
 
+#ifndef COMPAT_HAVE_VZALLOC
+static inline void *vzalloc(unsigned long size)
+{
+	void *rv = vmalloc(size);
+	if (rv)
+		memset(rv, 0, size);
+
+	return rv;
+}
+#endif
+
+#ifndef COMPAT_HAVE_UMH_WAIT_PROC
+/* On Jul 17 2007 with commit 86313c4 usermodehelper: Tidy up waiting,
+ * UMH_WAIT_PROC was added as an enum value of 1.
+ * On Mar 23 2012 with commit 9d944ef3 that got changed to a define of 2. */
+#define UMH_WAIT_PROC 1
+#endif
+
 /* see upstream commit 2d3854a37e8b767a51aba38ed6d22817b0631e33 */
 #if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,30)
 #ifndef cpumask_bits
@@ -553,11 +596,6 @@ static inline int backport_bitmap_parse(const char *buf, unsigned int buflen,
 }
 #endif
 
-#ifndef __CHECKER__
-# undef __cond_lock
-# define __cond_lock(x,c) (c)
-#endif
-
 #ifndef net_random
 #define random32 net_random
 #endif
@@ -612,14 +650,6 @@ static inline void blk_queue_max_segments(struct request_queue *q, unsigned shor
 }
 #endif
 
-#ifndef COMPAT_HAVE_BOOL_TYPE
-typedef _Bool                   bool;
-enum {
-	false = 0,
-	true = 1
-};
-#endif
-
 /* REQ_* and BIO_RW_* flags have been moved around in the tree,
  * and have finally been "merged" with
  * 7b6d91daee5cac6402186ff224c3af39d79f4a0e and
@@ -647,7 +677,6 @@ enum {
 				 * now equivalent to bi_rw */
 
 #define DRBD_REQ_SYNC		REQ_SYNC
-#define DRBD_REQ_UNPLUG		REQ_UNPLUG
 #define DRBD_REQ_FLUSH		REQ_FLUSH
 #define DRBD_REQ_FUA		REQ_FUA
 #define DRBD_REQ_DISCARD	REQ_DISCARD
@@ -662,6 +691,14 @@ enum {
 #define DRBD_REQ_HARDBARRIER	0
 #endif
 
+/* again: testing on this _inside_ the ifdef REQ_FLUSH,
+ * see 721a960 block: kill off REQ_UNPLUG */
+#ifdef REQ_UNPLUG
+#define DRBD_REQ_UNPLUG		REQ_UNPLUG
+#else
+#define DRBD_REQ_UNPLUG		0
+#endif
+
 #else				/* "older", and hopefully not
 				 * "partially backported" kernel */
 
@@ -770,6 +807,12 @@ DISCARD requests, as our queue does not announce QUEUE_FLAG_DISCARD yet.
 #define time_is_after_eq_jiffies(a) time_before_eq(jiffies, a)
 #endif
 
+#ifndef time_in_range
+#define time_in_range(a,b,c) \
+	(time_after_eq(a,b) && \
+	 time_before_eq(a,c))
+#endif
+
 #ifdef COMPAT_BIO_SPLIT_HAS_BIO_SPLIT_POOL_PARAMETER
 #define bio_split(bi, first_sectors) bio_split(bi, bio_split_pool, first_sectors)
 #endif
@@ -1099,4 +1142,91 @@ static inline long __must_check IS_ERR_OR_NULL(const void *ptr)
 }
 #endif
 
+#ifndef disk_to_dev
+/* disk_to_dev was introduced with 2.6.27. Before that the kobj was directly in gendisk */
+static inline struct kobject *drbd_kobj_of_disk(struct gendisk *disk)
+{
+	return &disk->kobj;
+}
+#else
+static inline struct kobject *drbd_kobj_of_disk(struct gendisk *disk)
+{
+	return &disk_to_dev(disk)->kobj;
+}
+#endif
+
+#ifndef ULLONG_MAX
+/* introduced in 2.6.18 */
+#define ULLONG_MAX (~0ULL)
+#endif
+
+#ifndef SK_CAN_REUSE
+/* This constant was introduced by Pavel Emelyanov <xemul@parallels.com> on
+   Thu Apr 19 03:39:36 2012 +0000. Before the release of linux-3.5
+   commit 4a17fd52 sock: Introduce named constants for sk_reuse */
+#define SK_CAN_REUSE   1
+#endif
+
+#ifndef COMPAT_HAVE_KREF_SUB
+static inline int kref_sub(struct kref *kref, unsigned int count,
+	     void (*release)(struct kref *kref))
+{
+	WARN_ON(release == NULL);
+
+	if (atomic_sub_and_test((int) count, &kref->refcount)) {
+		release(kref);
+		return 1;
+	}
+	return 0;
+}
+#endif
+
+#ifndef KOBJECT_CREATE_AND_ADD_EXPORTED
+struct kobject *kobject_create_and_add(const char *name, struct kobject *parent);
+int kobject_init_and_add(struct kobject *kobj, struct kobj_type *ktype,
+			 struct kobject *parent, const char *name);
+#endif
+
+#ifdef COMPAT_KMAP_ATOMIC_PAGE_ONLY
+/* see 980c19e3
+ * highmem: mark k[un]map_atomic() with two arguments as deprecated */
+#define drbd_kmap_atomic(page, km)	kmap_atomic(page)
+#define drbd_kunmap_atomic(addr, km)	kunmap_atomic(addr)
+#else
+#define drbd_kmap_atomic(page, km)	kmap_atomic(page, km)
+#define drbd_kunmap_atomic(addr, km)	kunmap_atomic(addr, km)
+#endif
+
+#ifdef COMPAT_HAVE_NETLINK_SKB_PARMS_PORTID
+#define NETLINK_CB_PORTID(skb) NETLINK_CB(skb).portid
+#else
+#define NETLINK_CB_PORTID(skb) NETLINK_CB(skb).pid
+#endif
+
+
+#ifndef COMPAT_HAVE_LIST_SPLICE_TAIL_INIT
+static inline void __backported_list_splice(const struct list_head *list,
+					    struct list_head *prev,
+					    struct list_head *next)
+{
+        struct list_head *first = list->next;
+        struct list_head *last = list->prev;
+
+        first->prev = prev;
+        prev->next = first;
+
+        last->next = next;
+        next->prev = last;
+}
+
+static inline void list_splice_tail_init(struct list_head *list,
+                                         struct list_head *head)
+{
+        if (!list_empty(list)) {
+		__backported_list_splice(list, head->prev, head);
+                INIT_LIST_HEAD(list);
+        }
+}
+#endif
+
 #endif
