@@ -32,10 +32,10 @@ extern const char *drbd_buildtag(void);
 
 /* End of external module for 2.6.33 stuff */
 
-#define REL_VERSION "8.4.1"
+#define REL_VERSION "8.4.3"
 #define API_VERSION 1
 #define PRO_VERSION_MIN 86
-#define PRO_VERSION_MAX 100
+#define PRO_VERSION_MAX 101
 
 #ifndef __CHECKER__   /* for a sparse run, we need all STATICs */
 #define DBG_ALL_SYMBOLS /* no static functs, improves quality of OOPS traces */
@@ -53,7 +53,9 @@ extern const char *drbd_buildtag(void);
  */
 
 /* Enable fault insertion code */
-#define DRBD_ENABLE_FAULTS
+#ifndef CONFIG_DRBD_FAULT_INJECTION
+#define CONFIG_DRBD_FAULT_INJECTION 1
+#endif
 
 #ifdef __KERNEL__
 #include "compat.h"
