@@ -277,13 +277,11 @@ void range_check(const enum range_checks what, const char *name,
 
 struct d_option *new_opt(char *name, char *value)
 {
-	struct d_option *cn = malloc(sizeof(struct d_option));
+	struct d_option *cn = calloc(1, sizeof(struct d_option));
 
 	/* fprintf(stderr,"%s:%d: %s = %s\n",config_file,line,name,value); */
 	cn->name = name;
 	cn->value = value;
-	cn->mentioned = 0;
-	cn->is_escaped = 0;
 
 	return cn;
 }
@@ -1674,9 +1672,23 @@ void set_on_hosts_in_res(struct d_resource *res)
 			/* Simple: host->on_hosts = concat_names(l_res->me->on_hosts, l_res->peer->on_hosts); */
 			last = NULL;
 			for (host2 = l_res->all_hosts; host2; host2 = host2->next)
-				if (!host2->lower_name)
+				if (!host2->lower_name) {
 					append_names(&host->on_hosts, &last, host2->on_hosts);
 
+					for_each_host(h, host2->on_hosts) {
+						struct d_volume *vol;
+
+						for_each_volume(vol, host->volumes)
+							check_uniq("device-minor", "device-minor:%s:%u", h->name,
+								   vol->device_minor);
+
+						for_each_volume(vol, host->volumes)
+							if (vol->device)
+								check_uniq("device", "device:%s:%s", h->name,
+									   vol->device);
+					}
+				}
+
 			host->lower = l_res;
 
 			/* */
@@ -1700,14 +1712,6 @@ void set_disk_in_res(struct d_resource *res)
 		if (!host->lower)
 			continue;
 
-		if (res->stacked && host->lower->stacked) {
-			fprintf(stderr,
-				"%s:%d: in resource %s, stacked-on-top-of %s { ... }:\n"
-				"\tFIXME. I won't stack stacked resources.\n",
-				res->config_file, res->start_line, res->name, host->lower_name);
-			config_valid = 0;
-		}
-
 		if (host->lower->ignore)
 			continue;
 
