@@ -216,7 +216,7 @@ struct option wait_cmds_options[] = {
 	{ "wfc-timeout",required_argument, 0, 't' },
 	{ "degr-wfc-timeout",required_argument,0,'d'},
 	{ "outdated-wfc-timeout",required_argument,0,'o'},
-	{ "wait-after-sb",no_argument,0,'w'},
+	{ "wait-after-sb",optional_argument,0,'w'},
 	{ 0,            0,           0,  0  }
 };
 
@@ -369,7 +369,7 @@ static const char *error_messages[] = {
 	"the kernel. (Maybe you need to modprobe it, or modprobe hmac?)",
 	EM(ERR_AUTH_ALG_ND) = "The 'cram-hmac-alg' you specified is not a digest.",
 	EM(ERR_NOMEM) = "kmalloc() failed. Out of memory?",
-	EM(ERR_DISCARD) = "--discard-my-data not allowed when primary.",
+	EM(ERR_DISCARD_IMPOSSIBLE) = "--discard-my-data not allowed when primary.",
 	EM(ERR_DISK_CONFIGURED) = "Device is attached to a disk (use detach first)",
 	EM(ERR_NET_CONFIGURED) = "Device has a net-config (use disconnect first)",
 	EM(ERR_MANDATORY_TAG) = "UnknownMandatoryTag",
@@ -1190,7 +1190,7 @@ int choose_timeout(struct choose_timo_ctx *ctx)
 		if (drbd_tla_parse(nlh)
 		|| timeout_parms_from_attrs(&parms, &info)) {
 			desc = "reply did not validate - "
-				"do you need to upgrade your useland tools?";
+				"do you need to upgrade your userland tools?";
 			goto error;
 		}
 		rr = parms.timeout_type;
@@ -1315,7 +1315,8 @@ static int generic_get_cmd(struct drbd_cmd *cm, int argc, char **argv)
 			break;
 
 		case 'w':
-			wait_after_split_brain = true;
+			if (!optarg || !strcmp(optarg, "yes"))
+				wait_after_split_brain = true;
 			break;
 
 		case 'D':
@@ -1473,7 +1474,7 @@ static int generic_get_cmd(struct drbd_cmd *cm, int argc, char **argv)
 				 * should continuous_poll continue?
 				 */
 				desc = "reply did not validate - "
-					"do you need to upgrade your useland tools?";
+					"do you need to upgrade your userland tools?";
 				rv = OTHER_ERROR;
 				goto out2;
 			}
@@ -1554,28 +1555,30 @@ static char *af_to_str(int af)
 
 static void show_address(void* address, int addr_len)
 {
-	struct sockaddr     *addr;
-	struct sockaddr_in  *addr4;
-	struct sockaddr_in6 *addr6;
+	union {
+		struct sockaddr     addr;
+		struct sockaddr_in  addr4;
+		struct sockaddr_in6 addr6;
+	} a;
 	char buffer[INET6_ADDRSTRLEN];
 
-	addr = (struct sockaddr *)address;
-	if (addr->sa_family == AF_INET
-	|| addr->sa_family == get_af_ssocks(0)
-	|| addr->sa_family == AF_INET_SDP) {
-		addr4 = (struct sockaddr_in *)address;
+	/* avoid alignment issues on certain platforms (e.g. armel) */
+	memset(&a, 0, sizeof(a));
+	memcpy(&a.addr, address, addr_len);
+	if (a.addr.sa_family == AF_INET
+	|| a.addr.sa_family == get_af_ssocks(0)
+	|| a.addr.sa_family == AF_INET_SDP) {
 		printI("address\t\t\t%s %s:%d;\n",
-		       af_to_str(addr4->sin_family),
-		       inet_ntoa(addr4->sin_addr),
-		       ntohs(addr4->sin_port));
-	} else if (addr->sa_family == AF_INET6) {
-		addr6 = (struct sockaddr_in6 *)address;
+		       af_to_str(a.addr4.sin_family),
+		       inet_ntoa(a.addr4.sin_addr),
+		       ntohs(a.addr4.sin_port));
+	} else if (a.addr.sa_family == AF_INET6) {
 		printI("address\t\t\t%s [%s]:%d;\n",
-		       af_to_str(addr6->sin6_family),
-		       inet_ntop(addr6->sin6_family, &addr6->sin6_addr, buffer, INET6_ADDRSTRLEN),
-		       ntohs(addr6->sin6_port));
+		       af_to_str(a.addr6.sin6_family),
+		       inet_ntop(a.addr6.sin6_family, &a.addr6.sin6_addr, buffer, INET6_ADDRSTRLEN),
+		       ntohs(a.addr6.sin6_port));
 	} else {
-		printI("address\t\t\t[unknown af=%d, len=%d]\n", addr->sa_family, addr_len);
+		printI("address\t\t\t[unknown af=%d, len=%d]\n", a.addr.sa_family, addr_len);
 	}
 }
 
